from kraken.lib.native import compile_and_load, I, A1D, A2D

lstm_utils = r"""
#include <math.h>

void sigmoid(int n,double v[n],double out[n]) {
    for(int i=0;i<n;i++) {
        double x = v[i];
        x = x<-100?-100:x>100?100:x;
        out[i] = 1.0/(1.0+exp(-x));
    }
}

void dotplus(int n,int m,double v[n],double a[n][m],double u[m]) {
    for(int i=0;i<n;i++) {
        double total;
        for(int j=0;j<m;j++) total += a[i][j]*u[j];
        v[i] += total;
    }
}

void prodplus(int n,double u[n],double v[n],double out[n]) {
    for(int i=0;i<n;i++) {
        out[i] += u[i]*v[i];
    }
}

void sumouter(int r,int n,int m,double a[n][m],double u[r][n],double v[r][m]) {
    for(int i=0;i<n;i++) {
        for(int j=0;j<m;j++) {
            double total = 0.0;
            for(int k=0;k<r;k++) total += u[k][i]*v[k][j];
            a[i][j] = total;
        }
    }
}

void sumprod(int r,int n,double u[r][n],double v[r][n],double a[n]) {
    for(int i=0;i<n;i++) {
        double total = 0.0;
        for(int k=0;k<r;k++) total += u[k][i]*v[k][i];
        a[i] = total;
    }
}
"""

lstm_native = compile_and_load(lstm_utils)
lstm_native.sigmoid.argtypes = [I, A1D, A1D]
lstm_native.dotplus.argtypes = [I, I, A1D, A2D, A1D]
lstm_native.prodplus.argtypes = [I, A1D, A1D, A1D]
lstm_native.sumouter.argtypes = [I, I, I, A2D, A2D, A2D]
lstm_native.sumprod.argtypes = [I, I, A1D, A2D, A2D]


def sigmoid(u, out=None):
    assert u.shape == out.shape and len(u.shape) == 1
    lstm_native.sigmoid(len(u), u, out)
    return out


def prodplus(u, v, out=None):
    assert u.shape == v.shape and u.shape == out.shape and len(u.shape) == 1
    lstm_native.prodplus(len(out), out, u, v)
    return out


def sumouter(u, v, out=None):
    assert out.shape == u.shape[1:] + v.shape[1:] and u.shape[:1] == v.shape[:1]
    lstm_native.sumouter(u.shape[0], out.shape[0], out.shape[1], out, u, v)
    return out


def sumprod(u, v, out=None):
    assert out.shape == u.shape[1:] and out.shape == v.shape[1:] and u.shape[:1] == v.shape[:1]
    lstm_native.sumprod(len(u), len(out), out, u, v)
    return out



################################################################
### various add-ons to the SciPy morphology package
################################################################

from numpy import *
from scipy.ndimage import morphology,measurements,filters
from scipy.ndimage.morphology import *

def label(image,**kw):
    """Redefine the scipy.ndimage.measurements.label function to
    work with a wider range of data types.  The default function
    is inconsistent about the data types it accepts on different
    platforms."""
    try: return measurements.label(image,**kw)
    except: pass
    types = ["int32","uint32","int64","unit64","int16","uint16"]
    for t in types:
	try: return measurements.label(array(image,dtype=t),**kw) 
	except: pass
    # let it raise the same exception as before
    return measurements.label(image,**kw)

def find_objects(image,**kw):
    """Redefine the scipy.ndimage.measurements.find_objects function to
    work with a wider range of data types.  The default function
    is inconsistent about the data types it accepts on different
    platforms."""
    try: return measurements.find_objects(image,**kw)
    except: pass
    types = ["int32","uint32","int64","unit64","int16","uint16"]
    for t in types:
	try: return measurements.find_objects(array(image,dtype=t),**kw) 
	except: pass
    # let it raise the same exception as before
    return measurements.find_objects(image,**kw)
    
def check_binary(image):
    assert image.dtype=='B' or image.dtype=='i' or image.dtype==dtype('bool'),\
        "array should be binary, is %s %s"%(image.dtype,image.shape)
    assert amin(image)>=0 and amax(image)<=1,\
        "array should be binary, has values %g to %g"%(amin(image),amax(image))

def r_dilation(image,size,origin=0):
    """Dilation with rectangular structuring element using maximum_filter"""
    return filters.maximum_filter(image,size,origin=origin)

def r_erosion(image,size,origin=0):
    """Erosion with rectangular structuring element using maximum_filter"""
    return filters.minimum_filter(image,size,origin=origin)

def r_opening(image,size,origin=0):
    """Opening with rectangular structuring element using maximum/minimum filter"""
    check_binary(image)
    image = r_erosion(image,size,origin=origin)
    return r_dilation(image,size,origin=origin)

def r_closing(image,size,origin=0):
    """Closing with rectangular structuring element using maximum/minimum filter"""
    check_binary(image)
    image = r_dilation(image,size,origin=0)
    return r_erosion(image,size,origin=0)

def rb_dilation(image,size,origin=0):
    """Binary dilation using linear filters."""
    output = zeros(image.shape,'f')
    filters.uniform_filter(image,size,output=output,origin=origin,mode='constant',cval=0)
    return array(output>0,'i')

def rb_erosion(image,size,origin=0):
    """Binary erosion using linear filters."""
    output = zeros(image.shape,'f')
    filters.uniform_filter(image,size,output=output,origin=origin,mode='constant',cval=1)
    return array(output==1,'i')

def rb_opening(image,size,origin=0):
    """Binary opening using linear filters."""
    image = rb_erosion(image,size,origin=origin)
    return rb_dilation(image,size,origin=origin)

def rb_closing(image,size,origin=0):
    """Binary closing using linear filters."""
    image = rb_dilation(image,size,origin=origin)
    return rb_erosion(image,size,origin=origin)

def rg_dilation(image,size,origin=0):
    """Grayscale dilation with maximum/minimum filters."""
    return filters.maximum_filter(image,size,origin=origin)

def rg_erosion(image,size,origin=0):
    """Grayscale erosion with maximum/minimum filters."""
    return filters.minimum_filter(image,size,origin=origin)

def rg_opening(image,size,origin=0):
    """Grayscale opening with maximum/minimum filters."""
    image = r_erosion(image,size,origin=origin)
    return r_dilation(image,size,origin=origin)

def rg_closing(image,size,origin=0):
    """Grayscale closing with maximum/minimum filters."""
    image = r_dilation(image,size,origin=0)
    return r_erosion(image,size,origin=0)

def spread_labels(labels,maxdist=9999999):
    """Spread the given labels to the background"""
    distances,features = morphology.distance_transform_edt(labels==0,return_distances=1,return_indices=1)
    indexes = features[0]*labels.shape[1]+features[1]
    spread = labels.ravel()[indexes.ravel()].reshape(*labels.shape)
    spread *= (distances<maxdist)
    return spread

def keep_marked(image,markers):
    """Given a marker image, keep only the connected components
    that overlap the markers."""
    labels,_ = label(image)
    marked = unique(labels*(markers!=0))
    kept = in1d(labels.ravel(),marked)
    return (image!=0)*kept.reshape(*labels.shape)

def remove_marked(image,markers):
    """Given a marker image, remove all the connected components
    that overlap markers."""
    marked = keep_marked(image,markers)
    return image*(marked==0)

def correspondences(labels1,labels2):
    """Given two labeled images, compute an array giving the correspondences
    between labels in the two images."""
    q = 100000
    assert amin(labels1)>=0 and amin(labels2)>=0
    assert amax(labels2)<q
    combo = labels1*q+labels2
    result = unique(combo)
    result = array([result//q,result%q])
    return result

def propagate_labels_simple(regions,labels):
    """Given an image and a set of labels, apply the labels
    to all the regions in the image that overlap a label."""
    rlabels,_ = label(regions)
    cors = correspondences(rlabels,labels)
    outputs = zeros(amax(rlabels)+1,'i')
    for o,i in cors.T: outputs[o] = i
    outputs[0] = 0
    return outputs[rlabels]

def propagate_labels(image,labels,conflict=0):
    """Given an image and a set of labels, apply the labels
    to all the regions in the image that overlap a label.
    Assign the value `conflict` to any labels that have a conflict."""
    rlabels,_ = label(image)
    cors = correspondences(rlabels,labels)
    outputs = zeros(amax(rlabels)+1,'i')
    oops = -(1<<30)
    for o,i in cors.T:
        if outputs[o]!=0: outputs[o] = oops
        else: outputs[o] = i
    outputs[outputs==oops] = conflict
    outputs[0] = 0
    return outputs[rlabels]

def select_regions(binary,f,min=0,nbest=100000):
    """Given a scoring function f over slice tuples (as returned by
    find_objects), keeps at most nbest regions whose scores is higher
    than min."""
    labels,n = label(binary)
    objects = find_objects(labels)
    scores = [f(o) for o in objects]
    best = argsort(scores)
    keep = zeros(len(objects)+1,'i')
    for i in best[-nbest:]:
        if scores[i]<=min: continue
        keep[i+1] = 1
    # print scores,best[-nbest:],keep
    # print sorted(list(set(labels.ravel())))
    # print sorted(list(set(keep[labels].ravel())))
    return keep[labels]

def all_neighbors(image):
    """Given an image with labels, find all pairs of labels
    that are directly neighboring each other."""
    q = 100000
    assert amax(image)<q
    assert amin(image)>=0
    u = unique(q*image+roll(image,1,0))
    d = unique(q*image+roll(image,-1,0))
    l = unique(q*image+roll(image,1,1))
    r = unique(q*image+roll(image,-1,1))
    all = unique(r_[u,d,l,r])
    all = c_[all//q,all%q]
    all = unique(array([sorted(x) for x in all]))
    return all

################################################################
### Iterate through the regions of a color image.
################################################################

def renumber_labels_ordered(a,correspondence=0):
    """Renumber the labels of the input array in numerical order so
    that they are arranged from 1...N"""
    assert amin(a)>=0
    assert amax(a)<=2**25
    labels = sorted(unique(ravel(a)))
    renum = zeros(amax(labels)+1,dtype='i')
    renum[labels] = arange(len(labels),dtype='i')
    if correspondence:
        return renum[a],labels
    else:
        return renum[a]

def renumber_labels(a):
    """Alias for renumber_labels_ordered"""
    return renumber_labels_ordered(a)

def pyargsort(seq,cmp=cmp,key=lambda x:x):
    """Like numpy's argsort, but using the builtin Python sorting
    function.  Takes an optional cmp."""
    return sorted(range(len(seq)),key=lambda x:key(seq.__getitem__(x)),cmp=cmp)

def renumber_by_xcenter(seg):
    """Given a segmentation (as a color image), change the labels
    assigned to each region such that when the labels are considered
    in ascending sequence, the x-centers of their bounding boxes
    are non-decreasing.  This is used for sorting the components
    of a segmented text line into left-to-right reading order."""
    objects = [(slice(0,0),slice(0,0))]+find_objects(seg)
    def xc(o): 
        # if some labels of the segmentation are missing, we
        # return a very large xcenter, which will move them all
        # the way to the right (they don't show up in the final
        # segmentation anyway)
        if o is None: return 999999
        return mean((o[1].start,o[1].stop))
    xs = array([xc(o) for o in objects])
    order = argsort(xs)
    segmap = zeros(amax(seg)+1,'i')
    for i,j in enumerate(order): segmap[j] = i
    return segmap[seg]

def ordered_by_xcenter(seg):
    """Verify that the labels of a segmentation are ordered
    spatially (as determined by the x-center of their bounding
    boxes) in left-to-right reading order."""
    objects = [(slice(0,0),slice(0,0))]+find_objects(seg)
    def xc(o): return mean((o[1].start,o[1].stop))
    xs = array([xc(o) for o in objects])
    for i in range(1,len(xs)):
        if xs[i-1]>xs[i]: return 0
    return 1



import hashlib
import os
import time
import errno
import contextlib
import ctypes

from numpy.ctypeslib import ndpointer
from ctypes import c_int, c_float, c_double, c_byte

I = c_int
F = c_float
D = c_double
B = c_byte

for d in range(1, 4):
    for T, t in [("I", "int32"), ("F", "float32"),
                 ("D", "float64"), ("B", "int8"),
                 ("U", "uint8")]:
        exec "A%d%s = ndpointer(dtype='%s', ndim=%d, flags='CONTIGUOUS, ALIGNED')" % (d, T, t, d)


@contextlib.contextmanager
def lockfile(fname, delay=0.5):
    while 1:
        try:
            fd = os.open(fname, os.O_RDWR | os.O_CREAT | os.O_EXCL)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            time.sleep(delay)
            continue
        else:
            break
    try:
        yield fd
    finally:
        os.close(fd)
        os.unlink(fname)


class CompileError(Exception):
    pass


def compile_and_find(c_string, prefix=".pynative", opt="-g -O4", libs="-lm",
                     options="-shared -fopenmp -std=c99 -fPIC", verbose=0):
    if not os.path.exists(prefix):
        os.mkdir(prefix)
    m = hashlib.md5()
    m.update(c_string)
    base = m.hexdigest()
    if verbose:
        print "hash", base, "for", c_string[:20], "..."
    with lockfile(os.path.join(prefix, base + ".lock")):
        so = os.path.join(prefix, base+".so")
        if os.path.exists(so):
            if verbose:
                print "returning existing", so
            return so
        source = os.path.join(prefix, base+".c")
        with open(source, "w") as stream:
            stream.write(c_string)
        cmd = "gcc " + opt + " " + libs + " " + options + " " + source + " -o " + so
        if verbose:
            print "#", cmd
        if os.system(cmd) != 0:
            if verbose:
                print "compilation failed"
            raise CompileError()
        return so


def compile_and_load(c_string, **keys):
    path = compile_and_find(c_string, **keys)
    return ctypes.CDLL(path)
